package npac.util.Workers;

import java.util.Vector;
import java.util.Enumeration;

import npac.util.List;

import npac.util.Wakeup;

import npac.util.ListItem;

/**
* WorkerPool stores and/or produces worker threads. 
* Worker threads are recycled when they are no longer used <BR>
*
* @version 0.1.0 first draft, April 21, 1997
* @version 0.1   completed , August 11, 1997 (Hasan Timucin OZDEMIR)
*/
public class 
WorkerPool extends Thread 
{

  /** Number of Maximum Worker objects */
  int NumberOfWorkers = 50;
 
  /** Keeps the Worker Object in one List */
  Vector   WorkerList      = null;

  /** Keeps the available Worker Objects  */
  Vector   AvailableIndex  = null;

  /** Listeners */
  Vector listeners;
 
  /** Worker object's priority */
  int SlavePriority = 2; 

  /** Manager's priority */
  int MasterPriority = 4; 

  /** Growing percent of the Pool */
  float GrowPercentage = 0f;

  public static final String AVAILABLE = "Available";

  public static final String     BUSSY = "Bussy";

  npac.util.List BussyQueue  = null;

  /** The size of the time interval to swich processes */
  int timeSlice = 2000;

  private static boolean DEBUG = false;

  /**
  * Creates a minimum WorkerPool. 
  */
  public 
  WorkerPool() 
  {
     createObject();

  }// end of constructor

  /*
  * Creates a WorkerPool with given size. Note that if the 
  * given size is less then the minimum size of the Pool 
  * then it uses the minimum size.
  * @param _size The size of the WorkerPool 
  */
  public 
  WorkerPool(int _size)
  {
     NumberOfWorkers = Math.max(5, Math.min(NumberOfWorkers, _size) );

     createObject();

  }// end of constructor

  /*
  * Creates the necessary Worker Tables.
  */
  private void
  createObject()
  {
     WorkerList = new Vector(NumberOfWorkers);

     AvailableIndex = new Vector(NumberOfWorkers);
     
     /***********************************/
     /** Create an empty Worker objects */
     /***********************************/

     for(int i=0;i <NumberOfWorkers; i++)
         WorkerList.insertElementAt(new Worker(this), i);

     /*************************************/
     /** Create the Available List Object */
     /*************************************/

     for(int i=0;i <NumberOfWorkers; i++) 
         AvailableIndex.insertElementAt(AVAILABLE,i);  

     BussyQueue = new npac.util.List();

     listeners = new Vector(5);

  }// end of createObject()

  /**
  * Supplies a worker thread. A worker thread is picked from the pool, and
  * supplied to the caller. If no workers are available in the pool, 
  * then WorkerPoolFullException is going to be raised.
  * @return the activated worker thread, or null if none could be created
  */
  public synchronized Worker 
  getWorker() throws WorkerPoolFullException 
  {
    int WorkerIndex = AvailableIndex.indexOf(AVAILABLE);
 
    if(WorkerIndex>=0 && WorkerIndex<NumberOfWorkers)
    {

      AvailableIndex.setElementAt(BUSSY,WorkerIndex);

      try
      { 
        BussyQueue.addItem(new ListItem(WorkerList.elementAt(WorkerIndex)));

      } catch(Exception e) 
        {
          System.out.println(" Exception at getWorker : "+e.getMessage());
        };

      return (Worker) WorkerList.elementAt(WorkerIndex);
      
    } else throw new WorkerPoolFullException();

  }// end of getWorker()

  /**
  * Returns a worker thread to the pool of available workers. 
  * If the provided info about the Worker is not valid, then
  * it raises UnknownWorkerException.
  * @param worker the worker to be made available
  */
  public synchronized void 
  freeWorker(Worker worker) throws UnknownWorkerException 
  {
    int WorkerIndex = WorkerList.indexOf(worker);
   
    if(WorkerIndex>=0 && WorkerIndex<NumberOfWorkers)
    {

      try
      {  
         BussyQueue.removeItem(new ListItem(WorkerList.elementAt(WorkerIndex)));

      } catch(Exception e) 
        {
          System.out.println(" Exception at freeWorker : "+e.getMessage());
        };

      AvailableIndex.setElementAt(AVAILABLE,WorkerIndex);
      if(DEBUG) System.out.println(" Worker : "+WorkerIndex+" is free ");

      wakeUpEverybody();

    } else throw new UnknownWorkerException();

  }// end of freeWorker()

  /**
  * Prints the current status of the Worker objects in the pool.
  */
  public synchronized void
  printStatusOfWorkers()
  {
     int i;
     for(i=0;i<NumberOfWorkers;i++)
        if(((String)AvailableIndex.elementAt(i)).equals("Bussy"))
            System.out.println(i+" status : Bussy ");

  }// end of while

  /**
  * Returns the name of the component. 
  * @return "WorkerPool" - the name of WorkerPool
  */
  public String 
  getComponentName() 
  {
    return("WorkerPool");
  }// end of getComponentName()

  /**
  * Notifies the WorkerPool object.
  */
  public synchronized void 
  wakeup()
  { notify(); 
  }// end of wakeup()

  /**
  * Thread interface.
  */
  public synchronized void
  run()
  {
     ListItem currentListItem = null;
     Worker   current; 
     boolean there_is_no_thread = true;

     while( true )
     {
        there_is_no_thread = true;

        while( there_is_no_thread )
        {
           if( BussyQueue.getSize() > 0 )
           {  
              try
              { 
                 currentListItem = BussyQueue.peekItem();
             
                 there_is_no_thread = ( ((Worker) currentListItem.getObject()).isActionSet() == false );
                 //there_is_no_thread = true;

              } catch (Exception e) {}

           }// end of if
 
           
           if( there_is_no_thread )
           {
               try
               { wait(2*timeSlice);
               } catch(Exception e) 
                 {  System.out.println(" WorkerPool: I cant' wait() "+e.getMessage()); 
                 };

               if(DEBUG) System.out.println(" WorkerPool: I woke up");
               //if(DEBUG) printStatusOfWorkers();

           }// end of if

        }// end of while

        if(DEBUG) System.out.println(" WorkerPool: BussyQueue size :"+BussyQueue.getSize());

        boolean SetPriorityOK = true;

        while( currentListItem != null )
        {
           /**************************/
           /** Get the Worker object */
           /**************************/
           current = (Worker) currentListItem.getObject();
           SetPriorityOK = true;
        
           try 
           { 
              if(current.isActionSet()) current.setPriority(getMasterPriority());
              else SetPriorityOK = false;

           } catch(Exception e) 
             {
                try { freeWorker(current); }
                catch(Exception e2) {};

                SetPriorityOK = false;
             }  
  
           /******************************/
           /** If this guy has an action */
           /******************************/
           if(SetPriorityOK == true)
           {

             /**************************************/
             /** Give sometime for the new process */
             /**************************************/
             try
             {
                wait(timeSlice);

             } catch(InterruptedException ie) 
               {}


             /*****************************************/
             /** Decrease the priority of this thread */
             /*****************************************/
             try 
             { 
                current.setPriority(getSlavePriority());

             } catch(Exception e) 
               { 
                  try { freeWorker(current); }
                  catch(Exception e2) {};
               } 

             if(DEBUG) printStatusOfWorkers();

           }// end of if

           /*********************************************/
           /** Proceed to the next element on the queue */
           /*********************************************/

           currentListItem = currentListItem.getSuccessor();

        }// end of while

     }// end of while

  }// end of run()

  /**
  * Registers the parties that would like to be notified when
  * there is some available worker in the pool.
  */
  public void
  registerFree(Wakeup ele)
  { listeners.addElement(ele);
  }// end of registerFree()

  /**
  * Notifies everybody to tell that there is a new free worker in the pool
  * looking for an action to take care of.
  */
  public void
  wakeUpEverybody()
  {
     if(listeners.isEmpty() == false)
     {
        Enumeration eles = listeners.elements();

        while(eles.hasMoreElements() )
        {
           Wakeup sleepyhead = (Wakeup) eles.nextElement();
           sleepyhead.wakeUp();

        }// end of while()   

     }// end of if

  }// end of wakeUpEverybody()

  /**
  * Sets the priority of the Master Thread.
  * <br>
  * This should be between 3 and 10.
  * @param a The priority of the Master Thread.
  *
  */
  public void
  setMasterPriority(int a)
  {
     if(a>2 && a<11 && (a-getSlavePriority()) > 1) MasterPriority = a;
  }// end of setMasterPriority()

  /**
  * @return The priority of the Master Thread.
  */
  public int
  getMasterPriority()
  {
      return MasterPriority;
  }// end of getMasterPriority()

  /**
  * Sets the priority of the Worker Threads.
  * <br>
  * This should be between 3 and 10.
  * @param a The priority of the Worker Thread.
  */
  public void
  setSlavePriority(int b)
  {
      if(b>0 && b<9 && (getMasterPriority()-b)>1 ) SlavePriority = b;
  }// end of setSlavePriority()

  /**
  * @return The priority of the Worker Threads.
  */
  public int
  getSlavePriority()
  {
      return SlavePriority;
  }// end of getSlavePriority()

  /**
  * Sets the growing percentage of the WorkerPool.
  * <br>
  * This should be between 0.0 and 0.5.
  * @param _growPercent Growing percentage of the WorkerPool.
  */
  public void
  setGrowPercent(float _growPercent)
  {
     if(_growPercent>=0 && _growPercent<=0.5)
          GrowPercentage = _growPercent;
  }// end of setGrowPercent()

  /**
  */
  public float
  getGrowPercent()
  {
     return GrowPercentage;
  }// end of getGrowPercent()

  /**
  *
  * @return Can this WorkerPool increase its Worker number when it is
  *         necessary? 
  */
  public boolean
  isGrowable()
  {
      return (GrowPercentage>0);
  }// end of isGrowable()

  /**
  * Sets the time slice between processes.
  * <br>
  * This value should be between 1 and 1000000.
  * @param _slice The size of the time interval.
  */
  public void
  setTimeSlice(long _slice)
  {
  }// end of setTimeSlice()
  
  
  /**
  * @return the time slice.
  */
  public long
  getTimeSlice()
  { return timeSlice;
  }// end of getTimeSlice()

}// end of WorkerPool



